import sys
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
import dask.array as da

######################
### NEEDED MODULES ###
######################

# there are only available in MN5 (load them in the same order)
"""
module purge
module load netcdf4-python/1.6.5-foss-2023b
module load xarray/2024.5.0-gfbf-2023b-Python-3.11.5
module load dask/2024.5.1-gfbf-2023b-Python-3.11.5
module load cfgrib/0.9.15.0-foss-2023b-Python-3.11.5
module load ecCodes/2.35.1-gompi-2023b
"""


def _grib2transport(grib_file, var_name, mesh_file, e12_mesh, e3_mesh):
    """
    Convert grib file to netCDF

    Parameters
    ----------
    grib_file: str or pathlib.Path
        Path to the file to be converted with the data in the original
        grid from Nemo.
    save_path: str or pathlib.Path
        Path to save the netcdf file.
    var_name: str
        Name of the variable in the grid_file, should be defined in
        var_defs.
    mesh_file: str or pathlib.Path
        Path to the Nemo mesh file wich contins level, latitude, and
        longitude data.

    Returns
    -------
    None

    """
    ###################
    ### GET VAR DEF ###
    ###################

    # grid shape
    y_len, x_len = 3606, 4322

    #################
    ### OPEN DATA ###
    #################

    ds = xr.open_mfdataset(grib_file, engine="cfgrib")

    ################
    ### FIX TIME ###
    ################

    # rename time coordinate
    ds = ds.rename({"time": "start_time"})

    # get the (time, step) selection fro the average, 1st day of the next month
    valid_time = pd.DatetimeIndex(ds['valid_time'].values.ravel())
    valid_time_sel = (valid_time.day == 1).reshape(ds['valid_time'].shape)

    # get index to extract (time, step) combinations
    step_sel = np.where(valid_time_sel)[-1]
    time_sel = range(len(ds.start_time))

    # define DataArrays for advanced indexing
    # https://docs.xarray.dev/en/stable/user-guide/indexing.html#more-advanced-indexing
    start_time_idx = xr.DataArray(time_sel, dims='t')
    step_idx = xr.DataArray(step_sel, dims='t')

    # select (time, step) combinations
    ds = ds.isel(start_time=start_time_idx, step=step_idx)

    # get start_time, end_time, and avg_time
    start_time = ds['start_time']
    end_time = ds['valid_time']
    time = start_time+(end_time-start_time)*0.5

    time_bnds = xr.DataArray(
        np.transpose([start_time, end_time]),
        dims=["time", "bnds"]
    )

    # time and time_bnds are used later to build the final dataset

    #################
    ### OPEN AREA ###
    #################

    with xr.open_mfdataset(mesh_file).isel(t=0) as file:
        e12 = file[e12_mesh].astype("float32", copy=False).data
        e3 = file[e3_mesh].astype("float32", copy=False).data

    ################
    ### FIX GRID ###
    ################

    # get data
    darray = ds[var_name]

    # reshape data
    data = darray.data.reshape(*darray.shape[:-1], y_len, x_len)

    #################
    ### TRANSPORT ###
    #################

    # compute transport
    transport = data * e12 * e3

    # convert regular array to masked array
    transport = da.ma.masked_where(np.isnan(transport), transport)

    # vertical integration
    transport = transport.sum(axis=1)

    return dict(data=transport, time=time.data, time_bnds=time_bnds)


def grib2barot(vgrib_file, ugrib_file, save_path, vvar_name, uvar_name,
               mesh_file):
    """
    Convert grib file to netCDF

    Parameters
    ----------
    grib_file: str or pathlib.Path
        Path to the file to be converted with the data in the original
        grid from Nemo.
    save_path: str or pathlib.Path
        Path to save the netcdf file.
    var_name: str
        Name of the variable in the grid_file, should be defined in
        var_defs.
    mesh_file: str or pathlib.Path
        Path to the Nemo mesh file wich contins level, latitude, and
        longitude data.

    Returns
    -------
    None

    """
    vt = _grib2transport(vgrib_file, vvar_name, mesh_file, 'e1v', 'e3v_0')
    ut = _grib2transport(ugrib_file, uvar_name, mesh_file, 'e2u', 'e3u_0')
    vt_dt = vt['data'].compute()
    ut_dt = ut['data'].compute()

    # compute psi
    # cumsum y-axis
    psi = ut_dt.cumsum(axis=-2)
    # invert cumsum x-axis
    psi += vt_dt[..., ::-1].cumsum(axis=-1)[..., ::-1]
    psi *= -0.5
    # remove sum on last x row of x axis
    psi += 0.5*ut_dt[..., -1].sum(axis=-1)[..., None, None]
    # convert masked array to numpy.nan
    psi = da.ma.filled(psi, np.nan)

    with xr.open_mfdataset(mesh_file).isel(t=0) as file:
        lon = file['glamt'].astype('float32', copy=False).data
        lat = file['gphit'].astype('float32', copy=False).data

    da_stf = xr.DataArray(psi, dims=['time', 'j', 'i'])
    da_stf = da_stf.assign_coords(
        time=('time', vt['time']),
        j=('j', np.arange(1, da_stf.shape[1]+1)),
        i=('i', np.arange(1, da_stf.shape[2]+1)),
        lat=(['j', 'i'], lat),
        lon=(['j', 'i'], lon),
    )

    # assign attributes
    da_stf.time.attrs = dict(
        standard_name='time',
        long_name='time',
        axis='T',
        bounds='time_bnds'
    )
    da_stf.lon.attrs = dict(
        standard_name="longitude",
        units="degrees_east"
    )
    da_stf.lat.attrs = dict(
        standard_name="latitude",
        units="degrees_north"
    )
    da_stf.j.attrs = dict(
        long_name="cell index along second dimension",
        units="1"
    )
    da_stf.i.attrs = dict(
        long_name="cell index along first dimension",
        units="1"
    )
    da_stf.attrs = dict(
        standard_name='',
        long_name='Ocean Barotropic Volume Streamfunction',
        comment='Streamfunction or its approximation for free surface '
                'models. See OMDP document for details.',
        units='m3 s-1'
    )

    # create dataset
    ds_stf = xr.Dataset({'vsftbarot': da_stf, 'time_bnds':  vt['time_bnds']})


    #################
    ### SAVE FILE ###
    #################

    # define encoding
    def_enconding = {'_FillValue': None, 'dtype': 'float32'}
    encoding = {ivar: def_enconding for ivar in ds_stf.data_vars}
    encoding.update({ivar: def_enconding for ivar in ds_stf.coords})

    # fill value for the var
    encoding['vsftbarot']['missing_value'] = 1e20
    encoding['vsftbarot']['_FillValue'] = 1e20

    # define time encoding as attribute
    del encoding['time']
    del encoding['time_bnds']

    ds_stf.time.encoding['units'] = 'days since 1950-01-01'
    ds_stf.time.encoding['dtype'] = 'double'
    ds_stf.time.encoding['_FillValue'] = None
    ds_stf.time_bnds.encoding['units'] = 'days since 1950-01-01'
    ds_stf.time_bnds.encoding['dtype'] = 'double'
    ds_stf.time_bnds.encoding['_FillValue'] = None

    # save data
    ds_stf.to_netcdf(save_path, unlimited_dims='time', encoding=encoding)


if __name__ == "__main__":

    path_vo = Path(sys.argv[1])
    path_uo = (path_vo.parent.parent / 'uo' /
               path_vo.stem.replace("vo_", "uo_")).with_suffix(".grib")
    save_path = (path_vo.parent.parent / 'vsftbarot' /
                 path_vo.stem.replace("vo_", "vsftbarot_")).with_suffix(".nc")
    save_path.parent.mkdir(exist_ok=True, parents=True)

    # Skip existing files
    if not save_path.exists():
        grib2barot(
            vgrib_file=path_vo,
            ugrib_file=path_uo,
            save_path=save_path,
            vvar_name="avg_vo",
            uvar_name="avg_uo",
            mesh_file="/gpfs/scratch/bsc32/bsc032050/ifsnemo_grib_post/"
                      "production/mesh_NEMO_EERIE.nc"
        )
        print("Saved %s" % save_path)
