import numpy as np
import xarray as xr
import pandas as pd

######################
### NEEDED MODULES ###
######################

# there are only available in MN5 (load them in the same order)
"""
module purge
module load netcdf4-python/1.6.5-foss-2023b
module load xarray/2024.5.0-gfbf-2023b-Python-3.11.5
module load dask/2024.5.1-gfbf-2023b-Python-3.11.5
module load cfgrib/0.9.15.0-foss-2023b-Python-3.11.5
module load ecCodes/2.35.1-gompi-2023b
"""

################
### var_defs ###
################

# this is from CMIP tables, except grid_type, which should be set based on
# the variable grid (t, u, v)
var_defs = {
    "avg_uo": {
        "out_name": "uo",  # if given, name to save the var in the netcdf
                           # insted of the original
        "grid_type": "u",  # t, u, v (depends on input variable)
        "standard_name": "sea_water_x_velocity",
        "units": "m s-1",
        "cell_methods": "time: mean",
        "long_name": "Sea Water X Velocity",
        "comment": "Prognostic x-ward velocity component resolved by the model.",
    },
    "avg_vo" : {
        "out_name": "vo",  # if given, name to save the var in the netcdf
                           # insted of the original
        "grid_type": "v",  # t, u, v (depends on input variable)
        "standard_name": "sea_water_y_velocity",
        "units": "m s-1",
        "cell_methods": "time: mean",
        "long_name": "Sea Water Y Velocity",
        "comment": "Prognostic y-ward velocity component resolved by the model.",

    },
}


def grib2netcdf(grib_file, save_path, var_name, mesh_file):
    """
    Convert grib file to netCDF

    Parameters
    ----------
    grib_file: str or pathlib.Path
        Path to the file to be converted with the data in the original
        grid from Nemo.
    save_path: str or pathlib.Path
        Path to save the netcdf file.
    var_name: str
        Name of the variable in the grid_file, should be defined in
        var_defs.
    mesh_file: str or pathlib.Path
        Path to the Nemo mesh file wich contins level, latitude, and
        longitude data.

    Returns
    -------
    None

    """
    ###################
    ### GET VAR DEF ###
    ###################

    var_def = var_defs[var_name]
    lev_dim = "oceanModelLayer"
    # data in mesh file
    lev_mesh = "nav_lev"
    lon_mesh = "glam%s" % var_def['grid_type']
    lat_mesh = "gphi%s" % var_def['grid_type']
    var_attrs = {
        attrs: value
        for attrs, value in var_def.items()
        if attrs not in ("out_name", "grid_type")
    }

    #################
    ### OPEN DATA ###
    #################

    ds = xr.open_mfdataset(grib_file, engine="cfgrib")

    ################
    ### FIX TIME ###
    ################

    # rename time coordinate
    ds = ds.rename({"time": "start_time"})

    # get the (time, step) selection fro the average, 1st day of the next month
    valid_time = pd.DatetimeIndex(ds['valid_time'].values.ravel())
    valid_time_sel = (valid_time.day == 1).reshape(ds['valid_time'].shape)

    # get index to extract (time, step) combinations
    step_sel = np.where(valid_time_sel)[-1]
    time_sel = range(len(ds.start_time))

    # define DataArrays for advanced indexing
    # https://docs.xarray.dev/en/stable/user-guide/indexing.html#more-advanced-indexing
    start_time_idx = xr.DataArray(time_sel, dims='t')
    step_idx = xr.DataArray(step_sel, dims='t')

    # select (time, step) combinations
    ds = ds.isel(start_time=start_time_idx, step=step_idx)

    # get start_time, end_time, and avg_time
    start_time = ds['start_time']
    end_time = ds['valid_time']
    time = start_time+(end_time-start_time)*0.5

    # assign time to be the avg_time and make it a dimension
    ds = ds.assign_coords({"time": time})
    ds = ds.swap_dims({"t": "time"})

    # define the time_bnds
    ds['time_bnds'] = xr.DataArray(
        np.transpose([start_time, end_time]),
        dims=["time", "bnds"]
    )

    # set attributes
    ds.time.attrs = dict(
        standard_name="time", long_name="time", axis="T", bounds="time_bnds"
    )

    # drop junk
    ds = ds.drop_vars(("start_time", "step", "valid_time"))

    #################
    ### OPEN MESH ###
    #################

    with xr.open_mfdataset(mesh_file).isel(t=0) as file:
        lev = file[lev_mesh].astype("float32", copy=False).data
        lon = file[lon_mesh].astype("float32", copy=False).data
        lat = file[lat_mesh].astype("float32", copy=False).data

    ################
    ### FIX GRID ###
    ################

    # get data
    da = ds[var_name]

    # reshape data
    data = da.data.reshape(*da.shape[:-1], *lon.shape)
    dims = list(da.dims[:-1]) + ["j", "i"]

    new_da = xr.DataArray(data, dims=dims)
    new_da = new_da.assign_coords(
        j=("j", np.arange(1, lon.shape[0]+1)),
        i=("i", np.arange(1, lon.shape[1]+1)),
        lat=(["j", "i"], lat),
        lon=(["j", "i"], lon),
    )

    # set attributes
    new_da.lon.attrs = dict(standard_name="longitude", units="degrees_east")
    new_da.lat.attrs = dict(standard_name="latitude", units="degrees_north")
    new_da.j.attrs = dict(long_name="cell index along second dimension",
                          units="1")
    new_da.i.attrs = dict(long_name="cell index along first dimension",
                          units="1")

    ds[var_name] = new_da

    ###############
    ### FIX LEV ###
    ###############

    # run this only for 3D data
    if lev_dim in ds.dims:
        # set coordinate
        ds = ds.assign_coords(lev=(lev_dim, lev))
        # set attributes
        ds.lev.attrs = dict(
            standard_name="depth",
            long_name="ocean depth coordinate",
            axis="Z",
            positive="down",
            units="m"
        )
        # swap dimension
        ds = ds.swap_dims({lev_dim: "lev"})
        # drop junk
        ds = ds.drop_vars((lev_dim,))

    ########################
    ### MODIFY VAR ATTRS ###
    ########################

    if "out_name" in var_def:
        # rename variable
        ds = ds.rename({var_name: var_def['out_name']})
        var_name = var_def['out_name']

    # set attributes
    ds[var_name].attrs = var_attrs

    #################
    ### SAVE FILE ###
    #################

    # define encoding
    def_enconding = {"_FillValue": None, "dtype": "float32"}
    encoding = {ivar: def_enconding for ivar in ds.data_vars}
    encoding.update({ivar: def_enconding for ivar in ds.coords})

    # fill value for the var
    encoding[var_name]["missing_value"] = 1e20
    encoding[var_name]["_FillValue"] = 1e20

    # define time encoding as attribute
    del encoding["time"]
    del encoding["time_bnds"]

    ds.time.encoding["units"] = "days since 1950-01-01"
    ds.time.encoding["dtype"] = "double"
    ds.time.encoding["_FillValue"] = None
    ds.time_bnds.encoding["units"] = "days since 1950-01-01"
    ds.time_bnds.encoding["dtype"] = "double"
    ds.time_bnds.encoding["_FillValue"] = None

    # save data
    ds.to_netcdf(save_path, unlimited_dims='time', encoding=encoding)


if __name__ == "__main__":

    grib2netcdf(
        grib_file="vo.grib",
        save_path="vo.nc",
        var_name="avg_vo",
        mesh_file="mesh_NEMO_EERIE.nc"
    )
